﻿using System;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.ComponentModel;

namespace CustomControls
{
    [DesignerAttribute(typeof(ListViewComboBoxDesigner))]
    public class ListViewComboBox : Control
    {
        private const int WM_SCROLL = 0x115;
        private const int KEY_PRESS_TIMEOUT = 500;
        private static Int32 m_DropDownButtonWidth = 20;
        private Label m_Label;
        private ExtendedListView m_ListView;
        private ToolStripDropDown m_ToolStripDropDown;
        private ToolStripControlHost m_ListViewHost;
        private FourStateListViewItem m_SelectedListViewItem;
        private System.Timers.Timer m_ToolStripDropDownTimer;
        private Boolean m_IsDropDown;
        private Boolean m_IsToolStripDropDownClosedTimerRunning;
        private Rectangle m_LabelRectangle;
        private Size m_DropDownWindowSize;
        private Size m_MinimumSize;
        private Int32 m_ToolStripDropDownClosedTimeCounter;
        private Int32 m_DefaultHeight;
        private Int32 m_CalculatedHeight;
        private Int32 m_ListViewIndex;
        private Int32 m_PreviousSelectedListViewItemIndex;
        private Color m_MouseOverListViewItemBackcolor;
        private Boolean m_ReadOnly;
        private Int32 m_MouseOverIndex;

        private System.Timers.Timer m_KeyPressTimer;
        private bool m_IsTimerRunning;
        private StringBuilder m_SearchString;

        public delegate void ChangedSelection(object sender, EventArgs e);
        public event ChangedSelection SelectionChanged;

        public delegate void DropDownHandler(object sender, EventArgs e);
        public event DropDownHandler DropDown;

        public ListViewComboBox()
        {
            //set styles
            this.SetStyle(ControlStyles.UserPaint, true);
            this.SetStyle(ControlStyles.DoubleBuffer, true);
            this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            this.SetStyle(ControlStyles.ResizeRedraw, true);

            //create instances
            this.m_ToolStripDropDownTimer = new System.Timers.Timer(50);
            this.m_Label = new Label();
            this.m_ListView = new ExtendedListView();
            this.m_ToolStripDropDown = new ToolStripDropDown();
            this.m_DropDownWindowSize = new Size(300, 200);
            this.PreviousSelectedListViewItemIndex = -1;
            this.SearchString = new StringBuilder();
            this.KeyPressTimer = new System.Timers.Timer();
            this.KeyPressTimer.AutoReset = true;
            this.KeyPressTimer.Interval = KEY_PRESS_TIMEOUT;
            this.KeyPressTimer.SynchronizingObject = this;
            this.KeyPressTimer.Elapsed += new System.Timers.ElapsedEventHandler(this.OnKeyPressTimer_Elapsed);

            //base control
            this.ReadOnly = false;
            this.DefaultHeight = 21;
            this.CalculatedHeight = this.DefaultHeight;
            this.Size = new Size(121, this.DefaultHeight);
            this.BackColor = System.Drawing.SystemColors.Window;
            this.LabelRectangle = new Rectangle(0, 0, this.Width - this.DropDownButtonWidth, this.Height - 4);
            this.MouseOverListViewItemBackcolor = Color.Orange;

            //handlers
            this.ListView.KeyPress += new KeyPressEventHandler(this.OnListView_KeyPress);
            this.m_ListView.Click += new EventHandler(this.OnListView_Click);
            this.m_ListView.ItemSelectionChanged += new ListViewItemSelectionChangedEventHandler(this.OnListView_ItemSelectionChanged);
            this.m_ListView.MouseMove += new MouseEventHandler(this.OnListView_MouseMove);
            this.m_ToolStripDropDown.Closed += new ToolStripDropDownClosedEventHandler(this.OnToolStripDropDown_Closed);
            this.m_ToolStripDropDownTimer.Elapsed += new System.Timers.ElapsedEventHandler(this.OnToolStripDropDownTimer_Elapsed);

            //child controls
            //set label
            this.m_Label.Size = new Size(LabelRectangle.Width, LabelRectangle.Height);
            this.m_Label.BorderStyle = BorderStyle.None;
            this.m_Label.Location = new Point(2, 2);
            this.m_Label.TextAlign = ContentAlignment.MiddleLeft;
            this.m_Label.BackColor = this.BackColor;
            this.m_Label.Text = null;
            this.Controls.Add(this.m_Label);

            //set listview
            this.m_ListView.FullRowSelect = true;
            this.m_ListView.GridLines = true;
            this.m_ListView.MultiSelect = false;
            this.m_ListView.Scrollable = true;
            this.m_ListView.HeaderStyle = ColumnHeaderStyle.Nonclickable;

            //set toolstripdowndown
            this.m_ToolStripDropDown.AutoSize = false;
            this.m_ToolStripDropDown.Margin = Padding.Empty;
            this.m_ToolStripDropDown.Padding = Padding.Empty;

            //set listviewhost and assign listview to listviewhost
            this.m_ListViewHost = new ToolStripControlHost(this.m_ListView);
            this.m_ListViewHost.AutoSize = false;
            this.m_ListViewHost.Margin = Padding.Empty;
            this.m_ListViewHost.Padding = Padding.Empty;

            //assign listviewhost to toolstripdowndown
            this.m_ToolStripDropDown.Items.Add(this.m_ListViewHost);

            //set initial size dropdown window size
            this.setDropDownWindowSize(this.DropDownWindowSize);
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            //suppress event in virtual mode to allow subitem search
            if (!this.ListView.VirtualMode)
            {
                base.OnKeyDown(e);
            }
        }

        [Browsable(true), DefaultValue(typeof(Color), "255,255,255"), Category("Appearance"), Description("The background color of the component.")]
        public override Color BackColor
        {
            get
            {
                return base.BackColor;
            }
            set
            {
                base.BackColor = value;
                this.Invalidate();
            }
        }

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
        [Browsable(true), DefaultValue(typeof(Size), "300,200"), Category("Misc"), Description("Size of dropdown window")]
        public Size DropDownWindowSize
        {
            get
            {
                return this.m_DropDownWindowSize;
            }
            set
            {
                this.m_DropDownWindowSize = value;
                this.setDropDownWindowSize(value);
                this.Invalidate();
            }
        }

        [Browsable(true), DefaultValue(typeof(Color), "255,165,0"), Category("Misc"), Description("Set backcolor for mouse move over listviewitem")]
        public Color MouseOverListViewItemBackcolor
        {
            get
            {
                return this.m_MouseOverListViewItemBackcolor;
            }
            set
            {
                this.m_MouseOverListViewItemBackcolor = value;
            }
        }

        [Browsable(false)]
        public new string Text
        {
            get
            {
                return this.m_Label.Text;
            }
            set
            {
                this.m_Label.Text = value;
            }
        }

        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
        public ExtendedListView ListView
        {
            get
            {
                return this.m_ListView;
            }
        }

        [Browsable(false)]
        public override Size MinimumSize
        {
            get
            {
                return this.m_MinimumSize;
            }
            set
            {
                this.m_MinimumSize = value;
            }
        }

        [Browsable(false)]
        public FourStateListViewItem SelectedListViewitem
        {
            get
            {
                return this.m_SelectedListViewItem;
            }
            private set
            {
                this.m_SelectedListViewItem = value;
                this.m_Label.Text = this.SelectedListViewitem.Text;
                if (this.SelectionChanged != null)
                {
                    this.SelectionChanged(this, EventArgs.Empty);
                }
            }
        }

        [Browsable(true), DefaultValue(false), Category("Misc"), Description("Controls whether the text in the edit control can be changed or not.")]
        public Boolean ReadOnly
        {
            get
            {
                return this.m_ReadOnly;
            }
            set
            {
                this.m_ReadOnly = value;
                if (this.BackColor.Equals(System.Drawing.SystemColors.Window))
                {
                    if (this.m_ReadOnly)
                    {
                        this.m_Label.BackColor = System.Drawing.SystemColors.Control;
                    }
                    else
                    {
                        this.m_Label.BackColor = System.Drawing.SystemColors.Window;
                    }
                }
            }
        }

        public string GetSearchString()
        {
            return this.SearchString.ToString();
        }

        private Int32 DropDownButtonWidth
        {
            get
            {
                return m_DropDownButtonWidth;
            }
        }

        private Rectangle LabelRectangle
        {
            get
            {
                return this.m_LabelRectangle;
            }
            set
            {
                this.m_LabelRectangle = value;
            }
        }

        private Boolean IsDropDown
        {
            get
            {
                return this.m_IsDropDown;
            }
            set
            {
                this.m_IsDropDown = value;
            }
        }

        private Boolean IsToolStripDropDownCloseTimerRunning
        {
            get
            {
                return this.m_IsToolStripDropDownClosedTimerRunning;
            }
            set
            {
                this.m_IsToolStripDropDownClosedTimerRunning = value;
            }
        }

        private Int32 DefaultHeight
        {
            get
            {
                return this.m_DefaultHeight;
            }
            set
            {
                this.m_DefaultHeight = value;
            }
        }

        private Int32 CalculatedHeight
        {
            get
            {
                return this.m_CalculatedHeight;
            }
            set
            {
                this.m_CalculatedHeight = value;
            }
        }

        private Int32 ListViewIndex
        {
            get
            {
                return this.m_ListViewIndex;
            }
            set
            {
                this.m_ListViewIndex = value;
            }
        }

        private Int32 PreviousSelectedListViewItemIndex
        {
            get
            {
                return this.m_PreviousSelectedListViewItemIndex;
            }
            set
            {
                this.m_PreviousSelectedListViewItemIndex = value;
            }
        }

        private System.Timers.Timer KeyPressTimer
        {
            get
            {
                return this.m_KeyPressTimer;
            }
            set
            {
                this.m_KeyPressTimer = value;
            }
        }

        private StringBuilder SearchString
        {
            get
            {
                return this.m_SearchString;
            }
            set
            {
                this.m_SearchString = value;
            }
        }

        private bool IsTimerRunning
        {
            get
            {
                return this.m_IsTimerRunning;
            }
            set
            {
                this.m_IsTimerRunning = value;
            }
        }

        private Int32 MouseOverIndex
        {
            get
            {
                return this.m_MouseOverIndex;
            }
            set
            {
                this.m_MouseOverIndex = value;
            }
        }

        private void setDropDownWindowSize(Size size)
        {
            this.m_ListViewHost.Width = size.Width;
            this.m_ListViewHost.Height = size.Height;
            this.m_ToolStripDropDown.Width = size.Width;
            this.m_ToolStripDropDown.Height = size.Height;
            this.m_ListView.Size = new Size(size.Width, size.Height);
        }

        private void CloseToolStripDropDown()
        {
            this.SelectedListViewitem = (FourStateListViewItem) this.m_ListView.Items[this.ListViewIndex];
            this.OnToolStripDropDown_Closed(this, new ToolStripDropDownClosedEventArgs(ToolStripDropDownCloseReason.CloseCalled));
        }

        protected override void OnFontChanged(EventArgs e)
        {
            Int32 newHeight = Convert.ToInt32(this.Font.Size * 2.54);
            if (newHeight > this.DefaultHeight)
            {
                this.CalculatedHeight = newHeight;
            }
            else
            {
                this.CalculatedHeight = this.DefaultHeight;
            }
            base.OnFontChanged(e);
            this.OnResize(e);
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);
            if (this.Size.Height != this.CalculatedHeight)
            {
                this.Size = new Size(this.Width, this.CalculatedHeight);
            }
            this.LabelRectangle = new Rectangle(0, 0, this.Width - this.DropDownButtonWidth, this.Height - 4);
            this.m_Label.Size = new Size(LabelRectangle.Width, LabelRectangle.Height);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);
            ControlPaint.DrawBorder3D(e.Graphics, this.ClientRectangle, Border3DStyle.Sunken);
            ControlPaint.DrawComboButton(e.Graphics, new Rectangle(this.Width - this.DropDownButtonWidth + 2, 2, this.DropDownButtonWidth - 4, this.Height - 4), ButtonState.Normal);
        }

        private void OnKeyPressTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            this.KeyPressTimer.Stop();
            this.IsTimerRunning = false;
            this.SearchString = new StringBuilder();
        }

        private void OnListView_KeyPress(object sender, KeyPressEventArgs e)
        {
            //timer
            if (this.IsTimerRunning)
            {
                this.KeyPressTimer.Interval = KEY_PRESS_TIMEOUT;
            }
            else
            {
                this.KeyPressTimer.Start();
                this.IsTimerRunning = true;
            }
            //check keychar 
            if (Convert.ToInt32(e.KeyChar) == 13) //enter
            {
                this.CloseToolStripDropDown();
            }
            //searchString
            this.SearchString.Append(e.KeyChar);
        }

        private void OnToolStripDropDown_Closed(object sender, ToolStripDropDownClosedEventArgs e)
        {
            this.m_ToolStripDropDown.Hide();
            this.IsToolStripDropDownCloseTimerRunning = true;
            this.m_ToolStripDropDownTimer.Start();
            if (this.PreviousSelectedListViewItemIndex > -1)
            {
                this.m_ListView.Items[this.PreviousSelectedListViewItemIndex].BackColor = this.m_ListView.BackColor;
            }
            this.PreviousSelectedListViewItemIndex = -1;
            this.IsDropDown = false;
        }

        protected override void OnClick(EventArgs e)
        {
            if (!this.IsDropDown && !this.IsToolStripDropDownCloseTimerRunning && !this.ReadOnly)
            {
                this.m_ToolStripDropDown.Show(this, 0, this.Height);
                this.ListView.Focus();
                this.IsDropDown = true;
                if (this.DropDown != null)
                {
                    this.DropDown(this, EventArgs.Empty);
                }
            }
        }

        private void OnListView_Click(Object sender, EventArgs e)
        {
            this.ListViewIndex = this.MouseOverIndex;
            this.CloseToolStripDropDown();
        }

        private void OnListView_ItemSelectionChanged(object sender, ListViewItemSelectionChangedEventArgs e)
        {
            this.ListViewIndex = e.ItemIndex;
            this.m_ListView.FocusedItem = this.m_ListView.Items[this.ListViewIndex];
        }

        private void OnListView_MouseMove(object sender, MouseEventArgs e)
        {
            ListViewItem listViewItem = null;
            if (this.ListView.Items.Count > 0)
            {
                Point p = new Point(e.X,e.Y);

                ListViewHitTestInfo listViewHitTestInfo = this.m_ListView.HitTest(p);
                listViewItem = listViewHitTestInfo.Item;
                if (listViewItem != null)
                {
                    this.MouseOverIndex = listViewItem.Index;
                    if (this.PreviousSelectedListViewItemIndex != listViewItem.Index)
                    {
                        if (this.PreviousSelectedListViewItemIndex > -1)
                        {
                            this.m_ListView.Items[this.PreviousSelectedListViewItemIndex].BackColor = this.m_ListView.BackColor;
                        }
                        this.PreviousSelectedListViewItemIndex = listViewItem.Index;
                        listViewItem.BackColor = MouseOverListViewItemBackcolor;
                    }
                }
            }
        }

        private void OnToolStripDropDownTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            if (this.m_ToolStripDropDownClosedTimeCounter < 2)
            {
                this.IsToolStripDropDownCloseTimerRunning = true;
                this.m_ToolStripDropDownClosedTimeCounter += 1;
            }
            else
            {
                this.m_ToolStripDropDownTimer.Stop();
                this.m_ToolStripDropDownClosedTimeCounter = 0;
                this.IsToolStripDropDownCloseTimerRunning = false;
            }
        }

        private void OnMouseScroll(ScrollEventArgs e)
        {
            if (e.Type == ScrollEventType.EndScroll)
            {
                this.Invalidate();
            }
        }

        
        protected override void WndProc(ref Message m)
        {
            base.WndProc(ref m);
            if (m.Msg.Equals(WM_SCROLL))
            {
                this.OnMouseScroll(new ScrollEventArgs((ScrollEventType)m.WParam.ToInt32(), 0));
            }
        }

    }
}
